/*
 *  ircd-hybrid: an advanced, lightweight Internet Relay Chat Daemon (ircd)
 *
 *  Copyright (c) 2004-2014 ircd-hybrid development team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 */

/*! \file m_etrace.c
 * \brief Includes required functions for processing the ETRACE command.
 * \version $Id$
 */

#include "stdinc.h"
#include "list.h"
#include "client.h"
#include "hash.h"
#include "irc_string.h"
#include "ircd.h"
#include "numeric.h"
#include "send.h"
#include "parse.h"
#include "modules.h"
#include "conf.h"
#include "conf_class.h"
#include "s_serv.h"


/* report_this_status()
 *
 * inputs       - pointer to client to report to
 *              - pointer to client to report about
 *              - flag full etrace or not
 * output       - NONE
 * side effects - NONE
 */
static void
report_this_status(struct Client *source_p, struct Client *target_p)
{
  if (target_p->status == STAT_CLIENT)
  {
    if (ConfigFileEntry.hide_spoof_ips)
      sendto_one(source_p, form_str(RPL_ETRACE),
                 me.name,
                 source_p->name,
                 HasUMode(target_p, UMODE_OPER) ? "Oper" : "User",
                 get_client_class(&target_p->localClient->confs),
                 target_p->name,
                 target_p->username,
                 target_p->realhost,
                 IsIPSpoof(target_p) ? "255.255.255.255" : target_p->sockhost,
                 target_p->localClient->client_host, target_p->localClient->client_server,
                 target_p->info);
    else
      sendto_one(source_p, form_str(RPL_ETRACE),
                 me.name,
                 source_p->name,
                 HasUMode(target_p, UMODE_OPER) ? "Oper" : "User",
                 get_client_class(&target_p->localClient->confs),
                 target_p->name,
                 target_p->username,
                 target_p->realhost,
                 target_p->sockhost,
                 target_p->localClient->client_host, target_p->localClient->client_server,
                 target_p->info);
  }
}

/*
 * do_etrace()
 */
static void
do_etrace(struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
  const char *tname = NULL;
  struct Client *target_p = NULL;
  int wilds = 0;
  int do_all = 0;
  dlink_node *ptr;

  if (parc > 2)
    if (hunt_server(client_p, source_p, ":%s ETRACE %s :%s", 2, parc, parv) != HUNTED_ISME)
      return;

  sendto_snomask(SNO_SPY, L_ALL,
                 "ETRACE requested by %s (%s@%s) [%s]",
                 source_p->name, source_p->username,
                 source_p->host, source_p->servptr->name);

  if (parc > 1)
  {
    tname = parv[1];

    if (tname != NULL)
      wilds = has_wildcards(tname);
    else
      tname = "*";
  }
  else
  {
    do_all = 1;
    tname = "*";
  }

  if (!wilds && !do_all)
  {
    target_p = hash_find_client(tname);

    if (target_p && MyClient(target_p))
      report_this_status(source_p, target_p);

    sendto_one(source_p, form_str(RPL_ENDOFTRACE), me.name,
               source_p->name, tname);
    return;
  }

  DLINK_FOREACH(ptr, local_client_list.head)
  {
    target_p = ptr->data;

    if (wilds)
    {
      if (!match(tname, target_p->name))
        report_this_status(source_p, target_p);
    }
    else
      report_this_status(source_p, target_p);
  }

  sendto_one(source_p, form_str(RPL_ENDOFTRACE), me.name,
             source_p->name, tname);
}

/* mo_etrace()
 *      parv[0] = sender prefix
 *      parv[1] = servername
 */
static void
mo_etrace(struct Client *client_p, struct Client *source_p,
          int parc, char *parv[])
{
  do_etrace(client_p, source_p, parc, parv);
}

static void
ms_etrace(struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
  do_etrace(client_p, source_p, parc, parv);
}

static struct Message etrace_msgtab =
{
  .cmd = "ETRACE",
  .args_max = MAXPARA,
  .handlers[UNREGISTERED_HANDLER] = m_unregistered,
  .handlers[CLIENT_HANDLER] = m_not_oper,
  .handlers[SERVER_HANDLER] = ms_etrace,
  .handlers[ENCAP_HANDLER] = m_ignore,
  .handlers[OPER_HANDLER] = mo_etrace,
};

static void
module_init(void)
{
  mod_add_cmd(&etrace_msgtab);
}

static void
module_exit(void)
{
  mod_del_cmd(&etrace_msgtab);
}

struct module module_entry =
{
  .version = "$Revision$",
  .modinit = module_init,
  .modexit = module_exit,
};
